--[[
Name: SanityBags-1.0
Revision: $Rev: 25000 $
Developed by: Antiarc
SVN: http://svn.wowace.com/root/trunk/SanityBags
Description: A powerful, customizable dynamic bags addon
Dependencies: Sanity2
License: MIT
Credits:
	# The Ace2 framework and Dewdrop libraries are the work of the most excellent Ace2 team
	# The frame sticking code is adapted from Bongos by Tuller, and uses FlyPaper, also by Tuller.
	# SanityBags is built on Sanity2, developed by Antiarc
	# Everything else developed by Antiarc
]]


SanityBags = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "AceHook-2.1", "AceDebug-2.0")
local AceOO = AceLibrary("AceOO-2.0")
local SML = AceLibrary("SharedMedia-1.0")

local dewdrop = AceLibrary("Dewdrop-2.0")

local L = AceLibrary("AceLocale-2.2"):new("SanityBags")

SanityBags:RegisterDB("SanityBagsDB", "SanityBagsDBPC")
SanityBags_SetDefaultDB()

local SanityBag = AceOO.Class("AceHook-2.1", "AceEvent-2.0")
local SanityBagItem = AceOO.Class("AceHook-2.1")
local qualities = {L["Trash"], L["Common"], L["Uncommon"], L["Rare"], L["Epic"], L["Legendary"]}

local bagSlotIDs = {0, 1, 2, 3, 4}
local bankSlotIDs = {-1, 5, 6, 7, 8, 9, 10, 11}

local attachPoints = {"TOPLEFT", "TOP", "TOPRIGHT", "RIGHT", "BOTTOMRIGHT", "BOTTOM", "BOTTOMLEFT", "LEFT"}

local TYPE_BAG = 1
local TYPE_BANK = 2

----------------------------------------------------------
-- Bag driver
----------------------------------------------------------

function SanityBags:OnInitialize()
	self.bags = {}
	self.bagsUID = {}
	self.handledItems = {}
	self.handleIndex = {}
	self.buttonPool = {}
	self.buttonCount = 0
	self.bagCount = 0
	
	self.ItemColors = {
		{157/255, 157/255, 157/255},
		{255/255, 255/255, 255/255},
		{30/255, 255/255, 0/255},
		{0/255, 112/255, 221/255},
		{163/255, 53/255, 238/255},
		{255/255, 128/255, 0/255}
	};
		
	-- self.bankInfo = SanityBagsBankInfo:new()
	
	for  idx, bag in pairs(self.db.char.bags) do
		local bag = self:CreateBag(bag)
	end
	
	-- Restore stuck frames
	for _, bag in ipairs(self.bags) do
		local frame = bag.frame
		local parent = self.bagsUID[bag.opts.anchor.puid]
		if parent and bag:CanAttachTo(parent) then
			-- self:Stick(bag)
			bag:Attach(parent, bag.opts.anchor.p, bag.opts.anchor.r, true)
		end
		bag:SetFilter(bag.opts.filter)
	end
	SanityBags:UpdateDefaultBags()
	SanityBags_RegisterChat()
end

function SanityBags:OnEnable()
	self:Hook("OpenAllBags", true)
	self:Hook("CloseAllBags", true)	
	self:Hook("ToggleBackpack", true)
	self:Hook("ToggleBag", true)
	self:RegisterEvent("ADDON_LOADED")
end

function SanityBags:ADDON_LOADED()
	for k, bag in pairs(self.bags) do
		bag:RegisterBinding(true)
	end
end

function SanityBags:getBagListOptionsTable()
	return {}
end

function SanityBags:OnDisable()
	self:Unhook("OpenAllBags", true)
	self:Unhook("CloseAllBags", true)	
	self:Unhook("ToggleBackpack", true)
	self:Unhook("ToggleBag", true)
	self:UnregisterEvent("ADDON_LOADED")
	for k,v in pairs(self.bags) do
		v.frame:Hide()
	end
end

function SanityBags:GetButton()
	if #self.buttonPool > 0 then
		return tremove(self.buttonPool, 1)
	else
		self.buttonCount = self.buttonCount + 1
		return SanityBagItem:new(self.buttonCount)
	end
	
end

function SanityBags:CreateBag(opts, bagType)
	self.bagCount = self.bagCount + 1
	if not opts then
		opts = {}
		tinsert(self.db.char.bags, opts)
	end
	if not opts.uid then
		self.db.char.uid_count = self.db.char.uid_count + 1
		opts.uid = self.db.char.uid_count
	end
	if not opts.bagType and bagType then
		opts.bagType = bagType
	end
	if self.bagsUID[opts.uid] then
		error("Bag with UID " .. opts.uid .. " already exists!")
	end
	local bag = SanityBag:new(self, self.bagCount, opts)	
	tinsert(self.bags, bag)
	self.bagsUID[opts.uid] = bag
	return bag
end

function SanityBags:DeleteBag(bag)
	local bagTypeCt = 0
	for k, v in pairs(self.bags) do
		if v.opts.bagType == bag.opts.bagType then
			bagTypeCt = bagTypeCt + 1
		end
	end
	if bagTypeCt < 2 then
		self:Print("This is your last " .. L["Bag Type " .. bag.opts.bagType] .. " bag. Unable to delete.")
	end
	for k,v in pairs(self.bags) do
		if v == bag then
			v:Clear()
			v.frame:Hide()
			v:Disable()
			for idx,opt in pairs(self.db.char.bags) do
				if opt == v.opts then
					tremove(self.db.char.bags, idx)
					break
				end
			end
			tremove(self.bags, k)
			break
		end		
	end	
end

function SanityBags:ToggleBackpack()
	for _, v in pairs(self.bags) do
		if v.opts.title:lower() == L["Backpack"]:lower() then
			if v.frame:IsVisible() then
				for _, v2 in pairs(self.bags) do
					v.frame:Hide()
				end
				return
			else
				v.frame:Show()
				return
			end
		end
	end
	if #self.bags > 0 then	
		local v = self.bags[1]
		if v.frame:IsVisible() then
			for _, v2 in pairs(self.bags) do		
				v2.frame:Hide()
			end
			return
		else
			v.frame:Show()
			return
		end
	end
end

function SanityBags:ToggleBag(id, force)
	if id > 4 or id < 0 then
		self.hooks.ToggleBag(id, force)
		return
	end
	if id > #self.bags then return end
	local v = self.bags[id + 1]
	if v.frame:IsVisible() and not force then
		v.frame:Hide()
		return
	else
		v.frame:Show()
		return
	end
end

function SanityBags:OpenAllBags()
	local hide = false
	if #self.bags > 0 and self.bags[1].frame:IsVisible() then hide = true end
	for _, v in pairs(self.bags) do
		if v.opts.open_with_all then
			if hide then
				v.frame:Hide()
			else
				v.frame:Show()
			end
		end
	end
end

function SanityBags:CloseAllBags()
	for _, v in pairs(self.bags) do
		if v.opts.open_with_all then
			v.frame:Hide()
		end
	end
end

function SanityBags:SetItemHandled(item, owner)
	if not item then return end
	if not self.handledItems[owner.id] then
		self.handledItems[owner.id] = {}
		SanityBags:Debug("Allocating table to handle item")
	end
	local idx = (item.Bag or -1) .. "_" .. (item.Slot or -1)
	if not self.handledItems[owner.id][idx] then
		self.handledItems[owner.id][idx] = true
		self.handleIndex[idx] = (self.handleIndex[idx] or 0) + 1
	end
end

function SanityBags:UnhandleItem(item, owner)
	if not item then return end
	if not self.handledItems[owner.id] then return end
	local idx = (item.Bag or -1) .. "_" .. (item.Slot or -1)
	if self.handledItems[owner.id][idx] then
		self.handledItems[owner.id][idx] = nil
		self.handleIndex[idx] = (self.handleIndex[idx] or 1) - 1
	end
end

function SanityBags:UnhandleItemsFromBag(bag, owner)
	if not self.handledItems[owner.id] then return end
	local size = 28
	-- if GetContainerNumSlots(KEYRING_CONTAINER) > size then size = GetContainerNumSlots(KEYRING_CONTAINER) end
	for i = 1, size do
		local idx = bag .. "_" .. i
		if self.handleIndex[idx] then
			self.handleIndex[idx] = self.handleIndex[idx] - 1
		end
		if self.handledItems[owner.id][idx] then
			self.handledItems[owner.id][idx] = nil
		end
	end
end

function SanityBags:ClearItemsHandled(owner)
	if not self.handledItems[owner.id] then return end
	for k, v in pairs(self.handledItems[owner.id]) do
		self.handleIndex[k] = self.handleIndex[k] - 1
		self.handledItems[owner.id][k] = nil
	end
end

function SanityBags:IsItemHandled(item, caller)
	local idx = (item.Bag or -1) .. "_" .. (item.Slot or -1)
	return self.handleIndex[idx] and self.handleIndex[idx] > 0 and ((self.handledItems[caller.id] and not self.handledItems[caller.id][idx]) or not self.handledItems[caller.id])
end

function SanityBags:UpdateDefaultBags()
	for _,v in ipairs(self.bags) do
		if v.opts.filter == "" then
			v:Refresh()
		end
	end
end

function SanityBags:Stick(bagObj)
	local minDist, cp, crp, b = -1, nil, nil, nil
	local C_SNAP_DISTANCE = 12 * 12
	local E_SNAP_DISTANCE = 16 * 16
	
	for _, bag in ipairs(self.bags) do
		if bag ~= bagObj and bagObj:CanAttachTo(bag) then
			local distance, p, rp = bagObj:AttachData(bag)
			if (distance < minDist or minDist == -1) and bagObj:CanAttachTo(bag) then
				minDist = distance
				cp = p
				crp = rp				
				b = bag
			end
		end
	end
	-- self:Print("Min distance is " .. minDist .. " (" .. attachPoints[cp] .. " from " .. attachPoints[crp] .. " on " .. b.frame.title:GetText())
	if b and ((cp % 2 == 0 and minDist < E_SNAP_DISTANCE) or (cp % 2 ~= 0 and minDist < C_SNAP_DISTANCE)) then
		if bagObj:Attach(b, cp, crp) then
				return
		end
	end
	bagObj:SetFloating()
end

function SanityBags:UpdateAnchor(obj)
	for k,v in pairs(self.bags) do
		local point,relativeTo,relativePoint,xOfs,yOfs = v.frame:GetPoint()
		if relativeTo == obj.frame then
			v:Attach(obj, point, relativePoint, true)
		end
	end
	local point,relativeTo,relativePoint,xOfs,yOfs = obj.frame:GetPoint(1)
	if relativeTo ~= UIParent and point then
		for k,v in pairs(self.bags) do
			if relativeTo == v.frame then
				obj:Attach(v, point, relativePoint, true)
				break
			end
		end
	end
end

----------------------------------------------------------
-- Bag prototype
----------------------------------------------------------

function SanityBag.prototype:init(parent, id, opts)
	SanityBag.super.prototype.init(self)

	self.buttons = {}
	self.items = {}
	self.filterResults = {}

	self.parent = parent
	self.id = id

	opts.scale = opts.scale or 1
	opts.filter = opts.filter or ""
	opts.search_string = opts.search_string or ""
	opts.title = opts.title or L["New Bag"]
	opts.itemSize = opts.itemSize or 32
	opts.itemsPerRow = opts.itemsPerRow or 4
	opts.bagType = opts.bagType or TYPE_BAG
	opts.color = opts.color or {r = 0.04, g = 0.05, b = 0, a = 0.7}
	opts.gorient = opts.gorient or "VERTICAL"
	opts.bcolor = opts.bcolor or {r = 0.8, g = 0.9, b = 1, a = 0.45}
	opts.anchorPoint = opts.anchorPoint or "BOTTOMRIGHT"
	opts.anchor = opts.anchor or {p = "CENTER", r = "CENTER", puid = 0, x = 0, y = 0}
	opts.sort = opts.sort or "Natural"
	opts.show_status_bar = opts.show_status_bar or false
	if opts.open_with_all == nil then
		opts.open_with_all = true
	end
	
	opts.showBagStatusTexture = opts.showBagStatusTexture or "Glaze"
	opts.showThisBagFree = opts.showThisBagFree or true
	opts.showBagStatusText = opts.showBagStatusText or true
	opts.showBagStatus = opts.showBagStatus or true	
	
	opts.gradamt = opts.gradamt or 0.6
	if opts.showSearch == nil then
		opts.showSearch = true
	end
	
	self.opts = opts
	
	-- Setup the frame	
	self.frame = CreateFrame("Frame", "SanityBag_" .. id, UIParent, "SanityBagFrameTemplate")
	self.frame:SetToplevel(true)

	self.frame:SetBackdropColor(self.opts.color.r, self.opts.color.g, self.opts.color.b, self.opts.color.a)
	
	self.frame:SetBackdropBorderColor(self.opts.bcolor.r, self.opts.bcolor.g, self.opts.bcolor.b, self.opts.bcolor.a)
	_G["SanityBag_" .. id .. "SearchContainer"]:SetBackdropBorderColor(self.opts.bcolor.r, self.opts.bcolor.g, self.opts.bcolor.b, self.opts.bcolor.a)
	_G["SanityBag_" .. id .. "FilterContainer"]:SetBackdropBorderColor(self.opts.bcolor.r, self.opts.bcolor.g, self.opts.bcolor.b, self.opts.bcolor.a)
	
	self.frame:SetClampedToScreen(true)
	self.frame.obj = self
	self.frame:SetScale(opts.scale or 1)
	self.frame.title = _G["SanityBag_" .. id .. "Title"]
	self.frame.title:SetText(opts.title)
	self.frame:Hide()
	self:SetWidth(self.opts.itemsPerRow)
	
	self.frame:SetScript("OnMouseDown", function()
		if arg1 == "LeftButton" and (not this.obj.opts.locked or IsControlKeyDown()) then
			this:StartMoving()
		end
		if arg1 == "RightButton" and IsControlKeyDown() then
			dewdrop:Open(this, "children", SanityBags_GetBagMenus(self), 'cursorX', true, 'cursorY', true)
		end
	end)
	
	_G["SanityBag_" .. id .. "FilterContainerSearchBox"]:SetText(opts.filter)
	_G["SanityBag_" .. id .. "SearchContainerSearchBox"]:SetText(opts.search_string)

	_G[self.frame:GetName() .. "FilterContainerSearchBox"]:SetScript("OnEnter", function() self:OnEnter() end)
	_G[self.frame:GetName() .. "FilterContainerSearchBox"]:SetScript("OnLeave", function() self:OnLeave() end)
	
	-- Hook stuff in.	
	self:Enable()
	self.lastFilter = "---------------"
	self:SetFilter(opts.filter, true)
	self:UpdateBarTexture()
	tinsert(UISpecialFrames, self.frame:GetName())
end

function SanityBag.prototype:Enable()
	self:Hook(SanityItemCache, "BAG_UPDATE")
	-- self:Hook("BankFrame_OnEvent", true)
	self:Hook(SanityItemCache, "AddItemTag")
	self:Hook(SanityItemCache, "RemoveItemTag")
	self:Hook(SanityItemCache, "SetItemCategory")
	self:Hook(SanityItemCache, "DeleteCategory")	
end

function SanityBag.prototype:Disable()
	self:Unhook(SanityItemCache, "BAG_UPDATE")
	-- self:Unhook("BankFrame_OnEvent")
	self:Unhook(SanityItemCache, "AddItemTag")
	self:Unhook(SanityItemCache, "RemoveItemTag")
	self:Unhook(SanityItemCache, "SetItemCategory")
	self:Unhook(SanityItemCache, "DeleteCategory")	
end

-- Hookses
----------------------------------------

function SanityBag.prototype:Refresh(force)
	self:SetFilter(self.opts.filter, force)
end

function SanityBag.prototype:AddItemTag(s, ...)
	self.hooks[s]["AddItemTag"](s, ...)
	self:Refresh(nil, true)
end

function SanityBag.prototype:RemoveItemTag(s, ...)
	self.hooks[s]["RemoveItemTag"](s, ...)
	self:Refresh(nil, true)
end

function SanityBag.prototype:SetItemCategory(s, ...)
	self.hooks[s]["SetItemCategory"](s, ...)
	self:Refresh(nil, true)
end

function SanityBag.prototype:DeleteCategory(s, ...)
	self.hooks[s]["DeleteCategory"](s, ...)
	self:Refresh(nil, true)
end

function SanityBag.prototype:BankFrame_OnEvent(event)
	if ( event == "BANKFRAME_OPENED" ) then
		self.bankVisible = true
		--[[
		if self.opts.bagType == TYPE_BANK then
			self.frame:Show()
		end
		]]--
	elseif ( event == "BANKFRAME_CLOSED" ) then
		self.bankVisible = false
		--[[
		if self.opts.bagType == TYPE_BANK then
			self.frame:Hide()
		end
		]]--
	end
	self.hooks["BankFrame_OnEvent"](event)
end

function SanityBag.prototype:BAG_UPDATE(s, arg1, ...)
	for k, v in pairs(self.items) do
		if v then
			SanityBags:UnhandleItemsFromBag(arg1, self)
		end
	end
	self.hooks[s]["BAG_UPDATE"](s, arg1, ...)
	self:ScheduleEvent("SanityBagRefresh" .. self.id, function() self:Refresh(); end, 0.1)
	self:UpdateBarFilled()
end

-- End hooks
----------------------------------------

function SanityBag.prototype:GetEmptySlots()
	local ii = 0
	local ji = 1
	return function()
		for i = ii, 4 do
			for j = ji, GetContainerNumSlots(i) do
				if not GetContainerItemInfo(i, j) then
					ii = i
					ji = i + 1
					return i, j
				end
			end
			ji = 1
		end
	end
end

function SanityBag.prototype:OnEnter()
	GameTooltip:SetOwner(this, "ANCHOR_LEFT")
	GameTooltip:ClearLines()
	for i = 1, 15 do
		GameTooltip:AddLine(L["filter help " .. i])
	end
	GameTooltip:Show()
end

function SanityBag.prototype:OnLeave()
	GameTooltip:Hide()
end

-- Frame anchoring manipulation
----------------------------------------

function SanityBag.prototype:CanAttachTo(bag)
	local frame = bag.frame
	for i = 1, frame:GetNumPoints() do
		local point,relativeTo,relativePoint,xOfs,yOfs = frame:GetPoint(i)
		if relativeTo == self.frame then
			return false
		else
			if relativeTo and relativeTo.obj then
				if not self:CanAttachTo(relativeTo.obj) then
					return false
				end
			end
		end
	end
	return true
end

function SanityBag.prototype:AttachData(bag)
	local function coord(b, p)
		local x,y = b.frame:GetCenter()
		local w,h = b.frame:GetWidth() / 2, b.frame:GetHeight() / 2

		if not x then
			SanityBags:Print("Error - no GetCenter(). This is a bug; please report it.")
			x = b.frame:GetLeft() + w
		end
		if not y then
			y = b.frame:GetTop() + h
		end
		local o = 0
		if p == 1 then return 		x - w, y + h
		elseif p == 2 then return	x - 0, y + h
		elseif p == 3 then return	x + w, y + h
		elseif p == 4 then return	x + w, y + 0
		elseif p == 5 then return	x + w, y - h
		elseif p == 6 then return	x + 0, y - h
		elseif p == 7 then return	x - w, y - h
		elseif p == 8 then return	x - w, y - 0
		end
		SanityBags:Print("Invalid index " .. p)
		return -1
	end
	
	local function distance(x1, y1, x2, y2)
		return ((x2 - x1) * (x2 - x1)) + ((y2 - y1) * (y2 - y1))
	end
	
	local minDist = -1
	local minPoint = -1
	local minRPoint = -1
	for i = 1, 7, 2 do
		local j = i + 2
		if j > 8 then j = j - 8 end
		local x1, y1 = coord(self, i)
		local x2, y2 = coord(bag, j)
		local d = distance(x1, y1, x2, y2)

		if minDist == -1 or d < minDist then minDist = d; minPoint = i; minRPoint = j end

		local j = i + 6
		if j > 8 then j = j - 8 end
		local x1, y1 = coord(self, i)
		local x2, y2 = coord(bag, j)
		local d = distance(x1, y1, x2, y2)
		if minDist == -1 or d < minDist then minDist = d; minPoint = i; minRPoint = j end
	end
	for i = 2, 8, 2 do
		local x1, y1 = coord(self, i)
		local j = i + 4
		if j > 8 then j = j - 8 end
		local x2, y2 = coord(bag, j)
		
		local d = distance(x1, y1, x2, y2)
		if minDist == -1 or d < minDist then minDist = d; minPoint = i; minRPoint = j end
	end
	
	return minDist, minPoint, minRPoint
end

function SanityBag.prototype:Attach(bag, attachPoint, relativePoint, noTranslate)
	if not self:CanAttachTo(bag) then return end
	if not noTranslate then
		attachPoint = attachPoints[attachPoint]
		relativePoint = attachPoints[relativePoint]
	end
	self.frame:ClearAllPoints()
	local offset = 0           
	SanityBags:Debug("Got attach point: " .. attachPoint)
	self.frame:SetPoint(attachPoint, bag.frame, relativePoint, 0, offset)
	self.frame:SetUserPlaced(false)
	
	self.opts.anchor.puid = bag.opts.uid
	self.opts.anchor.p = attachPoint
	self.opts.anchor.r = relativePoint
	self.opts.anchor.x = 0
	self.opts.anchor.y = offset

	do return true end
end

function SanityBag.prototype:UpdatePosition()
	SanityBags:Stick(self)
	local point, relativeTo, relativePoint, xOfs, yOfs = self.frame:GetPoint(1)
	self.opts.anchor.p = point
	self.opts.anchor.r = relativePoint
end

function SanityBag.prototype:ShowAttachments(show, stackDepth)
	if stackDepth > 10 then
		SanityBags:Print("Blew stack depth. Aborting...")
		return
	end
	if show then
		self.frame.bb_r,self.frame.bb_g,self.frame.bb_b,self.frame.bb_a = self.frame:GetBackdropBorderColor()	
		self.frame:SetBackdropBorderColor(0, 1, 0, 1)
	elseif self.frame.bb_a then
		self.frame:SetBackdropBorderColor(self.frame.bb_r, self.frame.bb_g, self.frame.bb_b, self.frame.bb_a	)
		self.frame.bb_a = nil
	else
		return
	end
	for k, v in pairs(SanityBags.bagsUID) do
		if v.opts.anchor.puid == self.opts.uid then
			v:ShowAttachments(show, stackDepth + 1)
		end
	end
end

function SanityBag.prototype:StickTo()
	local anchor = self.opts.anchor
	local parent = nil
	if anchor.puid > 0 and SanityBags.bagsUID[anchor.puid] then
		parent = SanityBags.bagsUID[anchor.puid].frame
	else
		parent = UIParent
	end

	self.frame:ClearAllPoints()
	local o = 0
	self.frame:SetPoint(anchor.p, parent, anchor.r, anchor.x, anchor.y + o)
end

function SanityBag.prototype:SetFloating()
	local anchor = self.opts.anchor
	if self.opts.anchor.puid == 0 then return end
	
	self.opts.anchor.puid = 0
	self.opts.anchor.p = self.opts.anchorPoint
	self.opts.anchor.r = self.opts.anchorPoint

	local ox, oy = self.frame:GetCenter()
	
	self.frame:ClearAllPoints()
	self.frame:SetPoint(anchor.p, UIParent, anchor.r, 0, 0)
	local nx, ny = self.frame:GetCenter()
	
	self.frame:ClearAllPoints()
	self.frame:SetPoint(anchor.p, parent, anchor.r, ox - nx, oy - ny)
end

-- Frame manipulation
----------------------------------------

function SanityBag.prototype:SetItemSize(size)
	for _, v in pairs(self.buttons) do
		v:SetSize(size)
	end
	self.opts.itemSize = size
	self:UpdateBounds()
end

function SanityBag.prototype:SetWidth(items)
	if not items then return end
	self.opts.itemsPerRow = items
	self:UpdateBounds()
	self:Arrange()
end

function SanityBag.prototype:UpdateBarTexture(v)
	_G[self.frame:GetName() .. "StatusBarBagFullFrameTexture"]:SetTexture(SML:Fetch("statusbar", self.opts.showBagStatusTexture))
	_G[self.frame:GetName() .. "StatusBarBagFullFrameBGTexture"]:SetTexture(SML:Fetch("statusbar", self.opts.showBagStatusTexture))
end

function SanityBag.prototype:UpdateBarFilled()
	local filled_slots, bag_slots = SanityItemCache:GetBagData(character)
	local div_slots = bag_slots
	if div_slots == 0 then div_slots = 1 end
	
	local tex = _G[self.frame:GetName() .. "StatusBarBagFullFrameBGTexture"]
	tex:SetVertexColor(tonumber(filled_slots) / tonumber(div_slots), 1 - (tonumber(filled_slots) / tonumber(div_slots)), 0, 0.4)

	local pw = tex:GetRight() - tex:GetLeft()
	local tex = _G[self.frame:GetName() .. "StatusBarBagFullFrameTexture"]
	tex:SetWidth(pw * (filled_slots / div_slots))
	tex:SetVertexColor(0, 1 - (tonumber(filled_slots) / tonumber(div_slots)), (tonumber(filled_slots) / tonumber(div_slots)) * 0.8)
	
	_G[self.frame:GetName() .. "StatusBarBagFullFrameText"]:SetText(filled_slots .. " / " .. bag_slots)
end

function SanityBag.prototype:Toggle(force)
	if self.frame:IsVisible() and not force then
		self.frame:Hide()
		return
	else
		self.frame:Show()
		return
	end
end

local function KeybindingClicked()
	this.bag:Toggle()
end

function SanityBag.prototype:RegisterBinding(silent)
	local set = self.opts.keybinding
	local frameName = self.frame:GetName() .. "_keybindingHandler"
	local frame = _G[frameName] or CreateFrame("Button", frameName, _G["UIParent"])
	if not set then
		frame:SetScript("OnMouseUp", nil)
	else
		frame.bag = self
		frame:SetScript("OnMouseUp", KeybindingClicked)
		SetBindingClick(set, frameName)
		if not silent then
			SanityBags:Print("Assigned " .. tostring(set) .. " to " .. self.opts.title)
		end
	end	
end

-- function SanityBag.prototype:Update

function SanityBag.prototype:UpdateBounds()
	local cols = self.opts.itemsPerRow
	local rows = math.ceil(#self.buttons / cols)
	if rows == 0 then rows = 1 end
	local pad = 0
	if self.opts.show_status_bar then
		_G[self.frame:GetName() .. "StatusBar"]:Show()
		pad = pad + 13
	else
		_G[self.frame:GetName() .. "StatusBar"]:Hide()
	end
	
	if(self.opts.showSearch) then
		_G[self.frame:GetName() .. "SearchContainer"]:SetHeight(23)
		_G[self.frame:GetName() .. "SearchContainer"]:Show()
		pad = pad + 21
	else
		_G[self.frame:GetName() .. "SearchContainer"]:SetHeight(1)
		_G[self.frame:GetName() .. "SearchContainer"]:Hide()
	end
	
	if(not self.opts.filter_hidden) then
		_G[self.frame:GetName() .. "FilterContainer"]:Show()
		pad = pad + 21
	else
		_G[self.frame:GetName() .. "FilterContainer"]:Hide()
	end	
	
	self.frame:SetWidth(cols * (self.opts.itemSize + 4) + 15)
	self.frame:SetHeight(rows * (self.opts.itemSize + 4) + 35 + pad)	
end

-- Item handling
----------------------------------------

function SanityBag.prototype:AddItem(item, claim)
	local button = SanityBags:GetButton()
	button:SetSize(self.opts.itemSize)
	button.slot:SetParent(self.frame)
	tinsert(self.buttons, button)
	if claim then
		SanityBags:SetItemHandled(item, self)
	end
	-- self.items[item] = 1
	button:SetItem(item)
	self:UpdateBounds()
end

function SanityBag.prototype:RemoveItem(item)
	for k, v in pairs(self.buttons) do
		if v.item == item then
			--self.items[item] = nil
			tremove(self.buttons, k):Free()
			SanityBags:UnhandleItem(item, self)
			self:UpdateBounds()
			break
		end
	end
end

function SanityBag.prototype:Clear()
	while #self.buttons > 0 do
		tremove(self.buttons):Free()
	end
	--[[
	for k,v in pairs(self.items) do
		self.items[k] = nil
	end
	]]--
end

function SanityBag.prototype:SetSearch(search, force)
	self.opts.search_string = search or self.opts.search_string
	self:ApplyFilter(self.opts.filter, force, self.opts.search_string)
end

function SanityBag.prototype:SetFilter(filter, force)	
	local diff = (filter ~= self.opts.filter) or force
	self.opts.filter = (filter or self.opts.filter):trim()
	self:ApplyFilter(self.opts.filter, diff, self.opts.search_string)	
end

function SanityBag.prototype:ApplyFilter(filter, force, namesearch)
	filter = filter or ""
	if filter ~= self.lastFilter or force then
		for k,v in pairs(self.filterResults) do self.filterResults[k] = nil end
	end
	self:Clear()
	SanityBags:ClearItemsHandled(self)
	local list = nil
	
	local idList = nil
	if self.opts.bagType == TYPE_BAG then
		if not SanityItemCache.db.realm.chars[UnitName("player")] then return end
		if not SanityItemCache.db.realm.chars[UnitName("player")].Bags then return end
		list = SanityItemCache.db.realm.chars[UnitName("player")].Bags
		idList = bagSlotIDs
	elseif self.opts.bagType == TYPE_BANK then
		if not self.bankVisible then
			self:Clear()
			SanityBags:ClearItemsHandled(self)
			self:Arrange()
			return
		end			
		if not SanityItemCache.db.realm.chars[UnitName("player")] then return end
		if not SanityItemCache.db.realm.chars[UnitName("player")].Bank then return end
		list = SanityItemCache.db.realm.chars[UnitName("player")].Bank
		idList = bankSlotIDs
	end
	if not list then return end
	
	local changed = false
	for _, index in ipairs(idList) do
		if list[index] then
			for idx, item in pairs(list[index]) do
				local added = false
				local handled = SanityBags:IsItemHandled(item, self)
				if not handled and #filter:trim() == 0 and item.Name and item.Name:lower():find(namesearch) then
					self:AddItem(item)
					added = true
					changed = true
				elseif #filter > 0 then
					local match = self.filterResults[item]
					if not match then
						match = SanityItemCache:MatchExpr(item, filter)
						self.filterResults[item] = 2
						if match then
							match = 1
						else
							match = 2
						end
						self.filterResults[item] = match
					end
					if match == 1 and item.Name and item.Name:lower():find(namesearch) then
						self:AddItem(item, filter ~= "*")
						added = true
						changed = true
					end
				end
				--[[
				if not added and self.items[item] then
					self:RemoveItem(item)
					changed = true
				end
				]]--
			end
		end
	end
	
	if #filter > 0 and filter ~= self.lastFilter then
		SanityBags:UpdateDefaultBags()
	end
	self.lastFilter = filter
	self:AddEmpty()
  	
	table.sort(self.buttons, function(a,b)
		if not a.item then return false
		elseif not b.item then return true
		elseif not a.item.Name then return false
		elseif not b.item.Name then return true
		end
		if self.opts.sort == "Natural" then
			if a.item.Bag == b.item.Bag then
				return a.item.Slot > b.item.Slot
			end
			return a.item.Bag > b.item.Bag
		else
			local e = (a.item[self.opts.sort] or "") == (b.item[self.opts.sort] or "")
			if e then return a.item.Name > b.item.Name end
			return (a.item[self.opts.sort] or "") > (b.item[self.opts.sort] or "")
		end
	end)
	self:Arrange()
end

function SanityBag.prototype:AddEmpty()
	local buttons = math.ceil(#self.buttons / self.opts.itemsPerRow) * self.opts.itemsPerRow
	if buttons == 0 then buttons = self.opts.itemsPerRow end 
	SanityBags:Debug("Bag " .. self.frame:GetName() .. " needs " .. buttons .. " buttons")
	local ct = 1

	for bag, slot in self:GetEmptySlots() do
		if #self.buttons >= buttons then break end
		local b = SanityBags:GetButton()
		b.slot:SetParent(self.frame)
		b:SetItem(nil)
		b:SetSlot(bag, slot)
		b:SetSize(self.opts.itemSize)
		tinsert(self.buttons, b)
	end
end

function SanityBag.prototype:Arrange()
	self.arranged = false
	if not self.frame:IsVisible() then return end
	self.arranged = true
	local ct = math.ceil(#self.buttons / self.opts.itemsPerRow) * self.opts.itemsPerRow
	local empty = 1
	
	local bpad = 0
	if self.opts.show_status_bar then
		bpad = 15
	end

	for j = 1, #self.buttons do
		local item = self.buttons[j]
		if j == 1 then
			item.slot:SetPoint("BOTTOMRIGHT", self.frame, "BOTTOMRIGHT", -10, 10 + bpad)
			corner = item.slot
		elseif ((j-1) % self.opts.itemsPerRow) == 0 then
			item.slot:SetPoint("BOTTOMRIGHT", corner, "TOPRIGHT", 0, 4)
			corner = item.slot
		else
			item.slot:SetPoint("BOTTOMRIGHT", last, "BOTTOMLEFT", -4, 0)
		end
		last = item.slot
	end
end

----------------------------------------------------------
-- Bag item prototype
----------------------------------------------------------

function SanityBagItem.prototype:init(id)
	SanityBagItem.super.prototype.init(self)
	local name = "SanityBagItem" .. id
	self.slot = CreateFrame("Button", name .. "Slot", UIParent)
	self.slot:SetWidth(16)
	self.slot:SetHeight(16)
	-- self.slot:SetFrameLevel(1)
	-- self.slot:SetNormalTexture("Interface\\Buttons\\UI-EmptySlot")
	-- self.slot:GetNormalTexture():SetTexCoord(0.1, 0.1, 0.9, 0.9)
	self.frame = CreateFrame("Button", name, self.slot, "ContainerFrameItemButtonTemplate")
	self.frame:SetPoint("TOPLEFT", self.slot, "TOPLEFT")
	self.frame:SetPoint("BOTTOMRIGHT", self.slot, "BOTTOMRIGHT")
	-- self.slot:SetFrameLevel(2)
	self.frame.tex = _G[name .. "IconTexture"]
	self.frame.tex:SetPoint("TOPLEFT", self.frame, "TOPLEFT")
	self.frame.tex:SetPoint("BOTTOMRIGHT", self.frame, "BOTTOMRIGHT")

	self.frame.normal_tex = self.frame:GetNormalTexture()
	self.frame.normal_tex:SetTexture("Interface\\Buttons\\UI-ActionButton-Border")
	self.frame.normal_tex:SetBlendMode("ADD")
	self.frame.normal_tex:SetAlpha(0.65)
	self.frame.normal_tex:SetPoint("CENTER", self.frame, "CENTER", 1, 0)
	self.frame.normal_tex:Show()

	self.frame.obj = self
	self.frame:Show()
	self.slot.frame = self.frame
	
	self.highlight = CreateFrame("Model", name .. "Highlighter", self.slot)
	self.highlight:Hide()
	self.highlight:SetModel("Interface\Buttons\UI-AutoCastButton.mdx")
	self.highlight:SetModelScale(1.2)
	self.highlight:SetAllPoints()
	self.highlight:SetSequence(0)
	self.highlight:SetSequenceTime(0, 0)

	self.slot:Hide()
end

function SanityBagItem.prototype:SetSize(size)
	self.slot:SetWidth(size)
	self.slot:SetHeight(size)
	self.frame.normal_tex:SetWidth(size*1.75)
	self.frame.normal_tex:SetHeight(size*1.75)
end

function SanityBagItem.prototype:Highlight()
	self.highlight:Show()
end

function SanityBagItem.prototype:Unhighlight()
	self.highlight:Hide()
end

function SanityBagItem.prototype:SetItem(item)
	self.item = item
	if item and item.Name then 
		self.slot:SetID(item.Bag)
		self.frame:SetID(item.Slot)
		local texture, itemCount, locked, quality, readable = GetContainerItemInfo(item.Bag, item.Slot)
		SetItemButtonTexture(self.frame, texture)
		SetItemButtonCount(self.frame, itemCount);
		SetItemButtonDesaturated(self.frame, locked, 0.5, 0.5, 0.5);
		if ( texture ) then
			self.frame.hasItem = 1
		else
			self.frame.hasItem = nil
		end
		local r = SanityBags.ItemColors[item.Quality +1]
		self.frame.normal_tex:SetVertexColor(r[1], r[2], r[3])
		self.frame.readable = readable
	else
		SetItemButtonCount(self.frame, 0);
		SetItemButtonDesaturated(self.frame, false, 0.5, 0.5, 0.5);
		SetItemButtonTexture(self.frame, nil)
		self.frame.normal_tex:SetVertexColor(0.4, 0.4, 0.4)
	end
	self.slot.frame:SetFrameLevel(self.slot:GetFrameLevel() + 2)
	self.slot:Show()
end

function SanityBagItem.prototype:SetSlot(bag, slot)
	self.slot:SetID(bag)
	self.frame:SetID(slot)
end

function SanityBagItem.prototype:Free()
	self.slot:ClearAllPoints()
	self.slot:SetPoint("CENTER", UIParent, "CENTER")
	self.slot:Hide()
	self.slot:SetID(-1)
	self.frame:SetID(-1)	
	SetItemButtonTexture(self.frame, nil)	
	tinsert(SanityBags.buttonPool, self)
end